# -*- coding: utf-8 -*-
"""
.. module: TODO
    :platform: TODO
    :synopsis: TODO

.. moduleauthor:: Aljosha Friemann a.friemann@automate.wtf
"""

import time
import uuid
import logging
import threading
import collections

from contextlib import contextmanager

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.ERROR)

logging._mdc = threading.local()  # type: ignore
start = time.time()


def get_mdc_fields():
    result = collections.defaultdict(None)
    contexts = vars(logging._mdc)  # type: ignore
    for context_id in sorted(contexts, key=lambda x: contexts[x].__creation_time__):
        result.update(**vars(contexts[context_id]))
    return result


@contextmanager
def new_log_context(**kwargs):
    context_id = "mdc-{thread}-{context}".format(
        thread=threading.current_thread().ident, context=uuid.uuid4()
    )

    LOGGER.debug("creating context %s", context_id)

    setattr(logging._mdc, context_id, threading.local())  # type: ignore

    context = getattr(logging._mdc, context_id)  # type: ignore

    context.__creation_time__ = time.time() - start

    for key, value in kwargs.items():
        setattr(context, key, value)

    try:

        yield context

    finally:

        LOGGER.debug("deleting context %s", context_id)

        try:
            delattr(logging._mdc, context_id)  # type: ignore
        except AttributeError:
            LOGGER.warning("context was already deleted %s", context_id)


MDContext = new_log_context



class ResourceBundleError(LookupError):
    pass


class MalformedResourceBundleError(ResourceBundleError):
    """
    Error that indicates that a ResourceBundle is malformed.
    """


class NotInResourceBundleError(ResourceBundleError):

    def __init__(self, bundle_name: str, key: str):
        """
        Error that is raised when a key could not be found in a ResourceBundle.

        :param str bundle_name: The name of the ResourceBundle
        :param str key: The key that could not be found
        """
        super(NotInResourceBundleError, self).__init__(f"Can't find key {key} in bundle {bundle_name}!")



import logging
from typing import List, Optional

import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype, is_period_dtype, is_string_dtype
from dateutil.relativedelta import relativedelta

from upgini.errors import ValidationError


class DateTimeSearchKeyConverter:
    def __init__(self, date_column: str, date_format: Optional[str] = None, logger: Optional[logging.Logger] = None):
        self.date_column = date_column
        self.date_format = date_format
        if logger is not None:
            self.logger = logger
        else:
            self.logger = logging.getLogger()
            self.logger.setLevel("FATAL")
        self.generated_features: List[str] = []

    @staticmethod
    def _int_to_opt(i: int) -> Optional[int]:
        if i == -9223372036855:
            return None
        else:
            return i

    def convert(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.copy()
        if is_string_dtype(df[self.date_column]):
            try:
                df[self.date_column] = pd.to_datetime(df[self.date_column], format=self.date_format)
            except ValueError as e:
                raise ValidationError(e)
        elif is_period_dtype(df[self.date_column]):
            df[self.date_column] = pd.to_datetime(df[self.date_column].astype("string"))
        elif is_numeric_dtype(df[self.date_column]):
            msg = f"Unsupported type of date column {self.date_column}. Convert to datetime please."
            self.logger.warning(msg)
            raise ValidationError(msg)

        # If column with date is datetime then extract seconds of the day and minute of the hour
        # as additional features
        seconds = "datetime_seconds"
        df[self.date_column] = df[self.date_column].dt.tz_localize(None)
        df[seconds] = (df[self.date_column] - df[self.date_column].dt.floor("D")).dt.seconds

        seconds_without_na = df[seconds].dropna()
        if (seconds_without_na != 0).any() and seconds_without_na.nunique() > 1:
            self.logger.info("Time found in date search key. Add extra features based on time")
            seconds_in_day = 60 * 60 * 24
            orders = [1, 2, 24, 48]
            for order in orders:
                sin_feature = f"datetime_time_sin_{order}"
                cos_feature = f"datetime_time_cos_{order}"
                df[sin_feature] = np.round(np.sin(2 * np.pi * order * df[seconds] / seconds_in_day), 10)
                df[cos_feature] = np.round(np.cos(2 * np.pi * order * df[seconds] / seconds_in_day), 10)
                self.generated_features.append(sin_feature)
                self.generated_features.append(cos_feature)

        df.drop(columns=seconds, inplace=True)

        df[self.date_column] = df[self.date_column].dt.floor("D").view(np.int64) // 1_000_000
        df[self.date_column] = df[self.date_column].apply(self._int_to_opt).astype("Int64")

        return df


def is_time_series(df: pd.DataFrame, date_col: str) -> bool:
    try:
        if df[date_col].isnull().any():
            return False

        df = pd.to_datetime(df[date_col]).to_frame()

        def rel(row):
            if not pd.isnull(row[date_col]) and not pd.isnull(row["shifted_date"]):
                return relativedelta(row[date_col], row["shifted_date"])

        value_counts = df[date_col].value_counts()
        # count with each date is constant
        if value_counts.nunique() == 1:
            # Univariate timeseries
            if value_counts.unique()[0] == 1:
                df["shifted_date"] = df[date_col].shift(1)
                # if dates cover full interval without gaps
                return df.apply(rel, axis=1).nunique() == 1

            # Multivariate timeseries
            df_with_unique_dates = df.drop_duplicates()

            df_with_unique_dates["shifted_date"] = df_with_unique_dates[date_col].shift(1)
            # if unique dates cover full interval without gaps
            return df_with_unique_dates.apply(rel, axis=1).nunique() == 1

        return False
    except Exception:
        return False
